from __future__ import absolute_import
from nlpaug.augmenter.word.word_augmenter import *
from nlpaug.augmenter.word.wordnet import *
from nlpaug.augmenter.word.random import *
from nlpaug.augmenter.word.word_embs import *
from nlpaug.augmenter.word.word_embs_aug import *
from nlpaug.augmenter.word.glove import *
from nlpaug.augmenter.word.word2vec import *
from nlpaug.augmenter.word.fasttext import *
from nlpaug.augmenter.word.bert import *
from nlpaug.augmenter.word.tfidf import *
from nlpaug.augmenter.word.spelling import *
from nlpaug.augmenter.word.stopwords import *


from __future__ import absolute_import
from nlpaug.augmenter.spectrogram.spectrogram_augmenter import *
from nlpaug.augmenter.spectrogram.frequency_masking import *
from nlpaug.augmenter.spectrogram.time_masking import *


"""
    Augmenter that apply pitch adjustment operation to audio.
"""

from nlpaug.augmenter.audio import AudioAugmenter
from nlpaug.util import Action
import nlpaug.model.audio as nma


class PitchAug(AudioAugmenter):
    """
    Augmenter that crop segment of audio by random values between crop_range variable.

    :param int sampling_rate: sampling rate of input audio
    :param tuple pitch_range: Range of applying pitch adjustment operation. Default value is (-10, 10)
        Number of half-steps that shifting audio
    :param str name: Name of this augmenter

    >>> import nlpaug.augmenter.audio as naa
    >>> aug = naa.PitchAug(sampling_rate=44010)
    """

    def __init__(self, sampling_rate, pitch_range=(-10, 10), name='Pitch_Aug', verbose=0):
        super(PitchAug, self).__init__(
            action=Action.SUBSTITUTE, name=name, verbose=verbose)
        self.model = self.get_model(sampling_rate, pitch_range)

    @classmethod
    def get_model(cls, sampling_rate, pitch_range):
        return nma.Pitch(sampling_rate, pitch_range)



import numpy as np

import nlpaug.util.math.normalization as normalization


class WordEmbeddings:
    def __init__(self, cache=True, skip_check=True):
        self.cache = cache
        self.skip_check = skip_check
        self.emb_size = 0
        self.vocab_size = 0
        self.embs = {}
        self.w2v = {}
        self.i2w = {}
        self.w2i = {}
        self.vectors = []
        self.normalized_vectors = None

        self.vocab = []

    def read(self, file_path, max_num_vector):
        raise NotImplementedError()

    def similar(self, word):
        raise NotImplementedError()

    def download(self, model_path):
        raise NotImplementedError()

    def word2idx(self, word):
        return self.w2i[word]

    def word2vector(self, word):
        return self.w2v[word]

    def idx2word(self, idx):
        return self.i2w[idx]

    def get_vectors(self, normalize=False):
        if normalize:
            return self.normalized_vectors
        return self.vectors

    def get_vocab(self):
        if self.cache:
            return self.vocab
        return [word for word in self.w2v]

    @classmethod
    def _normalize(cls, vectors, norm='l2'):
        if norm == 'l2':
            return normalization.l2_norm(vectors)
        elif norm == 'l1':
            return normalization.l1_norm(vectors)
        elif norm == 'standard':
            return normalization.standard_norm(vectors)

    def predict(self, word, top_n=10):
        source_id = self.word2idx(word)
        source_vector = self.word2vector(word)
        scores = np.dot(self.normalized_vectors, source_vector)
        target_ids = np.argpartition(-scores, top_n+1)[:top_n+1]
        target_words = [self.idx2word(idx) for idx in target_ids if idx != source_id]
        return target_words[:top_n]



from nlpaug.util import Method
from nlpaug import Augmenter
from nlpaug.util import WarningException, WarningName, WarningCode, WarningMessage


class CharAugmenter(Augmenter):
    def __init__(self, action, name='Char_Aug', aug_min=1, min_char=2, aug_char_p=0.3, aug_word_p=0.3,
                 tokenizer=None, reverse_tokenizer=None, stopwords=None,
                 verbose=0):
        super().__init__(
            name=name, method=Method.CHAR, action=action, aug_min=aug_min, verbose=verbose)
        self.aug_p = None
        self.aug_char_p = aug_char_p
        self.aug_word_p = aug_word_p
        self.min_char = min_char

        self.tokenizer = tokenizer or self._tokenizer
        self.reverse_tokenizer = reverse_tokenizer or self._reverse_tokenizer
        self.stopwords = stopwords

    @classmethod
    def _tokenizer(cls, text):
        return text.split(' ')

    @classmethod
    def token2char(cls, word):
        return list(word)

    @classmethod
    def _reverse_tokenizer(cls, tokens):
        return ' '.join(tokens)

    def skip_aug(self, token_idxes, tokens):
        return token_idxes

    def _get_aug_idxes(self, tokens, aug_p, mode):
        if mode == Method.CHAR:
            if len(tokens) <= self.min_char:
                return None

        aug_cnt = self.generate_aug_cnt(len(tokens), aug_p)
        idxes = [i for i, t in enumerate(tokens)]
        if mode == Method.WORD:
            if self.stopwords:
                idxes = [i for i in idxes if tokens[i] not in self.stopwords]
        elif mode == Method.CHAR:
            idxes = self.skip_aug(idxes, tokens)

        if len(idxes) == 0:
            if self.verbose > 0:
                exception = WarningException(name=WarningName.OUT_OF_VOCABULARY,
                                             code=WarningCode.WARNING_CODE_002, msg=WarningMessage.NO_WORD)
                exception.output()
            return None
        if len(idxes) < aug_cnt:
            aug_cnt = len(idxes)
        aug_idxes = self.sample(idxes, aug_cnt)
        return aug_idxes




"""
    Augmenter that apply typo error simulation to textual input.
"""

from nlpaug.augmenter.char import CharAugmenter
from nlpaug.util import Action, Method
import nlpaug.model.char as nmc


class KeyboardAug(CharAugmenter):
    """
    Augmenter that simulate typo error by random values. For example, people may type i as o incorrectly.\
        One keyboard distance is leveraged to replace character by possible keyboard error.

    :param int aug_min: Minimum number of character will be augmented.
    :param float aug_char_p: Percentage of character (per token) will be augmented.
    :param float aug_word_p: Percentage of word will be augmented.
    :param list stopwords: List of words which will be skipped from augment operation.
    :param func tokenizer: Customize tokenization process
    :param func reverse_tokenizer: Customize reverse of tokenization process
    :param str name: Name of this augmenter

    >>> import nlpaug.augmenter.char as nac
    >>> aug = nac.KeyboardAug()
    """

    def __init__(self, name='Keyboard_Aug', aug_min=1, aug_char_p=0.3, aug_word_p=0.3, stopwords=None,
                 tokenizer=None, reverse_tokenizer=None, verbose=0):
        super().__init__(
            action=Action.SUBSTITUTE, name=name, aug_char_p=aug_char_p, aug_word_p=aug_word_p, aug_min=aug_min,
            tokenizer=tokenizer, reverse_tokenizer=reverse_tokenizer, stopwords=stopwords, verbose=verbose)

        # TODO: support other type of keyboard
        self.keyboard_type = 'qwerty'
        self.model = self.get_model()

    def skip_aug(self, token_idxes, tokens):
        results = []
        for token_idx in token_idxes:
            # Some word does not come with vector. It will be excluded in lucky draw.
            char = tokens[token_idx]
            if char in self.model.model and len(self.model.predict(char)) > 1:
                results.append(token_idx)

        return results

    def substitute(self, data):
        results = []
        tokens = self.tokenizer(data)
        aug_word_idxes = self._get_aug_idxes(tokens, self.aug_word_p, Method.WORD)

        for token_i, token in enumerate(tokens):
            if token_i not in aug_word_idxes:
                results.append(token)
                continue

            result = ''
            chars = self.token2char(token)
            aug_char_idxes = self._get_aug_idxes(chars, self.aug_char_p, Method.CHAR)
            if aug_char_idxes is None:
                results.append(token)
                continue

            for char_i, char in enumerate(chars):
                if char_i not in aug_char_idxes:
                    result += char
                    continue

                result += self.sample(self.model.predict(chars[char_i]), 1)[0]

            results.append(result)

        return self.reverse_tokenizer(results)

    @classmethod
    def get_model(cls, include_spec=True, case_sensitive=True):
        return nmc.Keyboard(include_spec=include_spec, case_sensitive=case_sensitive)


"""
    Augmenter that apply mask operation to audio.
"""

from nlpaug.augmenter.audio import AudioAugmenter
from nlpaug.util import Action
import nlpaug.model.audio as nma


class MaskAug(AudioAugmenter):
    """
    Augmenter that crop segment of audio by random values between crop_range variable.

    :param int sampling_rate: sampling rate of input audio
    :param tuple mask_range: Range of applying mask operation. Default value is (0.2, 0.8)
        It means that first 20% and last 20% of data will not be excluded from augment operation. Augment operation
        will be applied to clip of rest of 60% time.
    :param int mask_factor: duration of masking period (in second)
    :param bool mask_with_noise: If it is True, targeting area will be replaced by noise. Otherwise, it will be
            replaced by 0.
    :param str name: Name of this augmenter

    >>> import nlpaug.augmenter.audio as naa
    >>> aug = naa.MaskAug(sampling_rate=44010)
    """

    def __init__(self, sampling_rate, mask_range=(0.2, 0.8), mask_factor=2, mask_with_noise=True,
                 name='Mask_Aug', verbose=0):
        super().__init__(
            action=Action.SUBSTITUTE, name=name, verbose=verbose)
        self.model = self.get_model(sampling_rate, mask_range, mask_factor, mask_with_noise)

    @classmethod
    def get_model(cls, sampling_rate, mask_range, mask_factor, mask_with_noise):
        return nma.Mask(sampling_rate, mask_range, mask_factor, mask_with_noise)



"""
    Augmenter that apply random character error to textual input.
"""

import string

from nlpaug.augmenter.char import CharAugmenter
from nlpaug.util import Action, Method


class RandomCharAug(CharAugmenter):
    """
    Augmenter that generate character error by random values. For example, people may type i as o incorrectly.

    :param str action: Possible values are 'insert', 'substitute', 'swap' and 'delete'. If value is 'insert', a new
        character will be injected to randomly. If value is 'substitute', a random character will be replaced
        original character randomly. If value is 'swap', adjacent characters within sample word will be swapped
        randomly. If value is 'delete', character will be removed randomly.
    :param int aug_min: Minimum number of character will be augmented.
    :param float aug_char_p: Percentage of character (per token) will be augmented.
    :param float aug_word_p: Percentage of word will be augmented.
    :param bool include_upper_case: If True, upper case character may be included in augmented data.
    :param bool include_lower_case: If True, lower case character may be included in augmented data.
    :param bool include_numeric: If True, numeric character may be included in augmented data.
    :param str spec_char: Special character may be included in augmented data.
    :param list stopwords: List of words which will be skipped from augment operation.
    :param func tokenizer: Customize tokenization process
    :param func reverse_tokenizer: Customize reverse of tokenization process
    :param str name: Name of this augmenter.

    >>> import nlpaug.augmenter.char as nac
    >>> aug = nac.QwertyAug()
    """

    def __init__(self, action=Action.SUBSTITUTE, name='RandomChar_Aug', aug_min=1, aug_char_p=0.3, aug_word_p=0.3,
                 include_upper_case=True, include_lower_case=True, include_numeric=True,
                 spec_char='!@#$%^&*()_+', stopwords=None, tokenizer=None, reverse_tokenizer=None, verbose=0):
        super().__init__(
            action=action, name=name, aug_char_p=aug_char_p, aug_word_p=aug_word_p, aug_min=aug_min,
            tokenizer=tokenizer, reverse_tokenizer=reverse_tokenizer, stopwords=stopwords, verbose=verbose)

        self.include_upper_case = include_upper_case
        self.include_lower_case = include_lower_case
        self.include_numeric = include_numeric
        self.spec_char = spec_char

        self.model = self.get_model()

    def insert(self, data):
        results = []
        tokens = self.tokenizer(data)
        aug_word_idxes = self._get_aug_idxes(tokens, self.aug_word_p, Method.WORD)

        for token_i, token in enumerate(tokens):
            if token_i not in aug_word_idxes:
                results.append(token)
                continue

            chars = self.token2char(token)
            aug_char_idxes = self._get_aug_idxes(chars, self.aug_char_p, Method.CHAR)
            if aug_char_idxes is None:
                results.append(token)
                continue

            aug_char_idxes.sort(reverse=True)
            for char_i in aug_char_idxes:
                chars.insert(char_i, self.sample(self.model, 1)[0])

            result = ''.join(chars)
            results.append(result)

        return self.reverse_tokenizer(results)

    def substitute(self, data):
        results = []
        tokens = self.tokenizer(data)
        aug_word_idxes = self._get_aug_idxes(tokens, self.aug_word_p, Method.WORD)

        for token_i, token in enumerate(tokens):
            if token_i not in aug_word_idxes:
                results.append(token)
                continue

            result = ''
            chars = self.token2char(token)
            aug_char_idxes = self._get_aug_idxes(chars, self.aug_char_p, Method.CHAR)
            if aug_char_idxes is None:
                results.append(token)
                continue

            for char_i, char in enumerate(chars):
                if char_i not in aug_char_idxes:
                    result += char
                    continue

                result += self.sample(self.model, 1)[0]

            results.append(result)

        return self.reverse_tokenizer(results)

    def swap(self, data):
        results = []
        tokens = self.tokenizer(data)
        aug_word_idxes = self._get_aug_idxes(tokens, self.aug_word_p, Method.WORD)

        for token_i, token in enumerate(tokens):
            if token_i not in aug_word_idxes:
                results.append(token)
                continue

            result = ''
            chars = self.token2char(token)
            original_chars = chars.copy()

            aug_char_idxes = self._get_aug_idxes(chars, self.aug_char_p, Method.CHAR)
            if aug_char_idxes is None:
                results.append(token)
                continue

            for char_i in aug_char_idxes:
                swap_position = self._get_swap_position(char_i, len(chars)-1)
                is_original_upper, is_swap_upper = chars[char_i].isupper(), chars[swap_position].isupper()
                chars[char_i], chars[swap_position] = original_chars[swap_position], original_chars[char_i]

                # Swap case
                if is_original_upper:
                    chars[char_i] = chars[char_i].upper()
                else:
                    chars[char_i] = chars[char_i].lower()
                if is_swap_upper:
                    chars[swap_position] = chars[swap_position].upper()
                else:
                    chars[swap_position] = chars[swap_position].lower()

                result += self.sample(self.model, 1)[0]

            result = ''.join(chars)
            results.append(result)

        return self.reverse_tokenizer(results)

    def delete(self, data):
        results = []
        tokens = self.tokenizer(data)
        aug_word_idxes = self._get_aug_idxes(tokens, self.aug_word_p, Method.WORD)

        for token_i, token in enumerate(tokens):
            if token_i not in aug_word_idxes:
                results.append(token)
                continue

            chars = self.token2char(token)
            aug_char_idxes = self._get_aug_idxes(chars, self.aug_char_p, Method.CHAR)
            if aug_char_idxes is None:
                results.append(token)
                continue

            aug_char_idxes.sort(reverse=True)
            for i in aug_char_idxes:
                del chars[i]

            result = ''.join(chars)
            results.append(result)

        return self.reverse_tokenizer(results)

    def get_model(self):
        candidates = []
        if self.include_upper_case:
            candidates += string.ascii_uppercase
        if self.include_lower_case:
            candidates += string.ascii_lowercase
        if self.include_numeric:
            candidates += string.digits
        candidates += self.spec_char

        return candidates

    def _get_swap_position(self, pos, token_length):
        if pos == 0:
            # Force swap with next character if it is first character
            return pos + 1
        elif pos == token_length:
            # Force swap with previous character if it is last character
            return pos - 1
        else:
            return pos + self.sample([-1, 1], 1)[0]




"""
    Augmenter that apply operation to textual input based on word embeddings.
"""

from nlpaug.augmenter.word import WordAugmenter
from nlpaug.util import Action


# @deprecated(deprecate_from='0.0.7', deprecate_to='0.0.9', msg="Use WordEmbsAug from 0.0.7 version")
class WordEmbsAugmenter(WordAugmenter):
    def __init__(self, model_path='.', action=Action.SUBSTITUTE,
                 name='WordEmbs_Aug', aug_min=1, aug_p=0.3, aug_n=5, n_gram_separator='_',
                 stopwords=None, tokenizer=None, reverse_tokenizer=None, verbose=0):
        super().__init__(
            action=action, name=name, aug_p=aug_p, aug_min=aug_min, stopwords=stopwords,
            tokenizer=tokenizer, reverse_tokenizer=reverse_tokenizer, verbose=verbose)
        self.model_path = model_path
        self.aug_n = aug_n
        self.model = self.get_model(force_reload=False)
        self.n_gram_separator = n_gram_separator

    def skip_aug(self, token_idxes, tokens):
        results = []
        for token_idx in token_idxes:
            # Some words do not come with vector. It will be excluded in lucky draw.
            word = tokens[token_idx]

            if word in self.model.w2v:
                results.append(token_idx)

        return results

    def insert(self, data):

        tokens = self.tokenizer(data)
        results = tokens.copy()

        aug_idexes = self._get_random_aug_idxes(tokens)
        if aug_idexes is None:
            return data
        aug_idexes.sort(reverse=True)

        for aug_idx in aug_idexes:
            new_word = self.sample(self.model.get_vocab(), 1)[0]
            if self.n_gram_separator in new_word:
                new_word = new_word.split(self.n_gram_separator)[0]
            results.insert(aug_idx, new_word)

        return self.reverse_tokenizer(results)

    def substitute(self, data):
        tokens = self.tokenizer(data)
        results = tokens.copy()

        aug_idexes = self._get_aug_idxes(tokens)
        if aug_idexes is None:
            return data

        for aug_idx in aug_idexes:
            original_word = results[aug_idx]
            candidate_words = self.model.predict(original_word, top_n=self.aug_n)
            substitute_word = self.sample(candidate_words, 1)[0]

            results[aug_idx] = substitute_word

        return self.reverse_tokenizer(results)



from nlpaug.util import Method
from nlpaug import Augmenter


class AudioAugmenter(Augmenter):
    def __init__(self, action, name='Audio_Aug', verbose=0):
        super(AudioAugmenter, self).__init__(
            name=name, method=Method.AUDIO, action=action, aug_min=1, verbose=verbose)

    def substitute(self, data):
        return self.model.manipulate(data)



import numpy as np
from nlpaug.model.word_embs import WordEmbeddings


class Fasttext(WordEmbeddings):
    def __init__(self, cache=True, skip_check=False):
        super().__init__(cache, skip_check)

    def read(self, file_path, max_num_vector=None):
        with open(file_path, 'r', encoding='utf-8') as f:
            header = f.readline()
            self.vocab_size, self.emb_size = map(int, header.split())

            for line in f:
                tokens = line.split()
                values = [val for val in tokens[(self.emb_size * -1):]]
                value_pos = line.find(' '.join(values))
                word = line[:value_pos-1]
                values = np.array([float(val) for val in values])

                self.vectors.append(values)
                self.i2w[len(self.i2w)] = word
                self.w2i[word] = len(self.w2i)
                self.w2v[word] = values

        self.vectors = np.asarray(self.vectors)
        if not self.skip_check:
            if len(self.vectors) != len(self.i2w):
                raise AssertionError('Vector Size:{}, Index2Word Size:{}'.format(len(self.vectors), len(self.i2w)))
            if len(self.i2w) != len(self.w2i):
                raise AssertionError('Index2Word Size:{}, Word2Index Size:{}'.format(len(self.i2w), len(self.w2i)))
            if len(self.w2i) != len(self.w2v):
                raise AssertionError('Word2Index Size:{}, Word2Vector Size:{}'.format(len(self.w2i), len(self.w2v)))

        self.normalized_vectors = self._normalize(self.vectors)

        if self.cache:
            self.vocab = [word for word in self.w2v]



# import numpy as np
#
# from nlpaug.model import Spectrogram
#
#
# class TimeWarping(Spectrogram):
#     def __init__(self, time_warp):
#         super(TimeWarping, self).__init__()
#
#         self.time_warp = time_warp
#
#     # TODO
#     def mask(self, mel_spectrogram):
#         """
#             From: https://arxiv.org/pdf/1904.08779.pdf,
#             Time warping is applied via the function
#             sparse image warp of tensorflow. Given
#             a log mel spectrogram with t time steps, we view it
#             as an image where the time axis is horizontal and the
#             frequency axis is vertical. A random point along the
#             horizontal line passing through the center of the image
#             within the time steps (W, t - W) is to be warped
#             either to the left or right by a distance w chosen from a
#             uniform distribution from 0 to the time warp parameter
#             W along that line.
#         :return:
#         """
#
#         time_range = mel_spectrogram.shape[1]
#         self.w = np.random.randint(self.time_warp)
#
#         center_point = np.random.randint(self.time_warp, time_range-self.time_warp)
#         distance = np.random.randint(-self.w, self.w)
#
#         # self.w0 = np.random.randint(time_range - self.t)
#         #
#         # augmented_mel_spectrogram = mel_spectrogram.copy()
#         # augmented_mel_spectrogram[:, self.time_warp:self.time_range-self.time_warp] = 0
#         # return augmented_mel_spectrogram
#         return mel_spectrogram



"""
    Augmenter that apply BERT's based operation to textual input.
"""

from nlpaug.augmenter.word import WordAugmenter
from nlpaug.util import Action
import nlpaug.model.lang_models as nml

BERT_MODEL = {}


def init_bert_model(model_path, device, tokenizer_path, force_reload=False):
    # Load model once at runtime

    global BERT_MODEL
    if BERT_MODEL and not force_reload:
        return BERT_MODEL

    bert_model = nml.Bert(model_path, tokenizer_path, device)
    bert_model.model.eval()
    BERT_MODEL = bert_model

    return bert_model


class BertAug(WordAugmenter):
    """
    Augmenter that leverage BERT's embeddings to find top n similar word for augmentation.

    :param str model_path: Model name or model path. It used pytorch-transformer to load the model.
    :param str tokenizer_path: Tokenizer name of path. It used pytorch-transformer to load the tokenizer.
    :param str action: Either 'insert or 'substitute'. If value is 'insert', a new word will be injected to random
        position according to contextual word embeddings calculation. If value is 'substitute', word will be replaced
        according to contextual embeddings calculation
    :param int aug_min: Minimum number of word will be augmented.
    :param float aug_p: Percentage of word will be augmented.
    :param int aug_n: Top n similar word for lucky draw
    :param list stopwords: List of words which will be skipped from augment operation.
    :param str name: Name of this augmenter

    >>> import nlpaug.augmenter.word as naw
    >>> aug = naw.BertAug()
    """

    def __init__(self, model_path='bert-base-uncased', tokenizer_path='bert-base-uncased', action=Action.SUBSTITUTE,
                 name='Bert_Aug', aug_min=1, aug_p=0.3, aug_n=5, stopwords=None, device='cpu', verbose=0):
        super().__init__(
            action=action, name=name, aug_p=aug_p, aug_min=aug_min, tokenizer=None, stopwords=stopwords,
            verbose=verbose)
        self.model_path = model_path
        self.tokenizer_path = tokenizer_path
        self.aug_n = aug_n
        self.device = device
        self.model = self.get_model(device=device, force_reload=False)
        self.tokenizer = self.model.tokenizer.tokenize
        self.reverse_tokenizer = self._reverse_tokenizer

    def skip_aug(self, token_idxes, tokens):
        results = []
        for token_idx in token_idxes:
            # Some token is not a partial word.
            word = tokens[token_idx]

            if word[:2] != nml.Bert.SUBWORD_PREFIX:
                results.append(token_idx)

        return results

    def _reverse_tokenizer(self, tokens):
        result = ''
        for token in tokens:
            if token[:2] == nml.Bert.SUBWORD_PREFIX:
                result += token[2:]
            else:
                result += ' ' + token
        return result[1:]

    def insert(self, data):
        tokens = self.tokenizer(data)
        results = tokens.copy()

        aug_idxes = self._get_random_aug_idxes(tokens)
        aug_idxes.sort(reverse=True)

        for aug_idx in aug_idxes:
            results.insert(aug_idx, nml.Bert.MASK)
            new_word = self.sample(self.model.predict(results, nml.Bert.MASK, self.aug_n), 1)[0]
            results[aug_idx] = new_word

        return self.reverse_tokenizer(results)

    def substitute(self, data):
        tokens = self.tokenizer(data)
        results = tokens.copy()

        aug_idxes = self._get_aug_idxes(tokens)

        for aug_idx in aug_idxes[:1]:
            original_word = results[aug_idx]
            candidate_words = self.model.predict(results, original_word, top_n=self.aug_n)
            substitute_word = self.sample(candidate_words, 1)[0]

            results[aug_idx] = substitute_word

        final_results = []
        for src, dest in zip(tokens, results):
            final_results.append(self.align_capitalization(src, dest))

        return self.reverse_tokenizer(final_results)

    def get_model(self, device='cpu', force_reload=False):
        return init_bert_model(self.model_path, device, self.tokenizer_path, force_reload)



"""
    Augmenter that apply time based masking to spectrogram input.
"""

from nlpaug.augmenter.spectrogram import SpectrogramAugmenter
from nlpaug.util import Action
import nlpaug.model.spectrogram as nms


class TimeMaskingAug(SpectrogramAugmenter):
    """
    Augmenter that mask spectrogram based on time by random values.

    :param int mask_factor: Value between 0 and mask_factor will be picked randomly.
        Mask range will be between [0, tau - master_factor) while tau is time range of input.
    :param str name: Name of this augmenter

    >>> import nlpaug.augmenter.spectogram as nas
    >>> aug = nas.TimeMaskingAug(mask_factor=80)
    """

    def __init__(self, mask_factor, name='TimeMasking_Aug', verbose=0):
        super(TimeMaskingAug, self).__init__(
            action=Action.SUBSTITUTE, name=name, aug_p=1, aug_min=0.3, verbose=verbose)

        self.model = self.get_model(mask_factor)

    def substitute(self, data):
        return self.model.mask(data)

    @classmethod
    def get_model(cls, mask_factor):
        return nms.TimeMasking(mask_factor)



import librosa
import numpy as np

from nlpaug.model.audio import Audio

"""
    Reference: https://www.kaggle.com/CVxTz/audio-data-augmentation
    A wrapper of librosa.effects.pitch_shift
"""

class Pitch(Audio):
    """
    Adjusting pitch

    :param sampling_rate: Sampling rate of input audio
    :param pitch_range: Number of half-steps that shifting audio
    """
    def __init__(self, sampling_rate, pitch_range):
        super(Pitch, self).__init__()

        self.sampling_rate = sampling_rate
        self.pitch_range = pitch_range

    def manipulate(self, data):
        n_step = np.random.randint(self.pitch_range[0], self.pitch_range[1])

        return librosa.effects.pitch_shift(data, self.sampling_rate, n_step)



"""
    Augmenter that apply random word operation to textual input.
"""

from nlpaug.augmenter.word import WordAugmenter
from nlpaug.util import Action


class RandomWordAug(WordAugmenter):
    """
    Augmenter that apply randomly behavior for augmentation.

    :param str action: Either 'swap' or 'delete'. If value is 'swap', adjacent words will be swapped randomly.
        If value is 'delete', word will be removed randomly.
    :param int aug_min: Minimum number of word will be augmented.
    :param float aug_p: Percentage of word will be augmented.
    :param list stopwords: List of words which will be skipped from augment operation.
    :param func tokenizer: Customize tokenization process
    :param func reverse_tokenizer: Customize reverse of tokenization process
    :param str name: Name of this augmenter

    >>> import nlpaug.augmenter.word as naw
    >>> aug = naw.RandomWordAug()
    """

    def __init__(self, action=Action.DELETE, name='RandomWord_Aug', aug_min=1, aug_p=0.3, stopwords=None,
                 tokenizer=None, reverse_tokenizer=None, verbose=0):
        super().__init__(
            action=action, name=name, aug_p=aug_p, aug_min=aug_min, stopwords=stopwords,
            tokenizer=tokenizer, reverse_tokenizer=reverse_tokenizer, verbose=verbose)

    def swap(self, data):
        results = self.tokenizer(data)
        aug_idxes = self._get_aug_idxes(results)
        original_tokens = results.copy()

        for i in aug_idxes:
            swap_position = self._get_swap_position(i, len(original_tokens) - 1)
            if len(results[i]) > 0:
                is_original_capitalize, is_swap_capitalize = results[i][0].isupper(), results[swap_position][0].isupper()
            else:
                is_original_capitalize = False
                is_swap_capitalize = False

            is_original_upper, is_swap_upper = results[i].isupper(), results[swap_position].isupper()
            results[i], results[swap_position] = original_tokens[swap_position], original_tokens[i]

            # Swap case
            if is_original_upper:
                results[i] = results[i].upper()
            elif is_original_capitalize:
                results[i] = results[i].capitalize()
            else:
                results[i] = results[i].lower()
            if is_swap_upper:
                results[swap_position] = results[swap_position].upper()
            elif is_swap_capitalize:
                results[swap_position] = results[swap_position].capitalize()
            else:
                results[swap_position] = results[swap_position].lower()

        return self.reverse_tokenizer(results)

    def _get_swap_position(self, pos, token_length):
        if pos == 0:
            # Force swap with next character if it is first character
            return pos + 1
        elif pos == token_length:
            # Force swap with previous character if it is last character
            return pos - 1
        else:
            return pos + self.sample([-1, 1], 1)[0]

    def delete(self, data):
        tokens = self.tokenizer(data)
        results = tokens.copy()

        aug_idxes = self._get_random_aug_idxes(tokens)
        aug_idxes.sort(reverse=True)

        for aug_idx in aug_idxes:
            del results[aug_idx]

        if len(results) > 0 and len(results[0]) > 0:
            results[0] = self.align_capitalization(tokens[0], results[0])

        return self.reverse_tokenizer(results)



class Method:
    CHAR = 'char'
    WORD = 'word'
    SPECTROGRAM = 'spectrogram'
    AUDIO = 'audio'

    FLOW = 'flow'

    @staticmethod
    def getall():
        return [Method.CHAR, Method.WORD, Method.AUDIO, Method.SPECTROGRAM, Method.FLOW]




"""
    Flow that apply augmentation sequentially.
"""

from nlpaug.util import Action
from nlpaug.flow import Pipeline


class Sequential(Pipeline):
    """
    Flow that apply augmenters sequentially.

    :param list flow: list of flow or augmenter
    :param str name: Name of this augmenter

    >>> import nlpaug.flow as naf
    >>> import nlpaug.augmenter.char as nac
    >>> import nlpaug.augmenter.word as naw
    >>> flow = naf.Sequential([nac.RandomCharAug(), naw.RandomWordAug()])
    """

    # TODO: Using epoch to implement 1-to-many
    def __init__(self, flow=None, name='Sequential_Pipeline', verbose=0):
        Pipeline.__init__(self, name=name, action=Action.SEQUENTIAL,
                          flow=flow, epoch=1, aug_min=-1, aug_p=1, verbose=verbose)

    def augment(self, data):
        """
        :param data: Data for augmentation
        :return: Augmented data

        >>> augmented_data = flow.augment(data)
        """
        results = []
        for _ in range(self.epoch):
            augmented_data = data[:]
            for aug in self:
                augmented_data = aug.augment(augmented_data)

            results.append(augmented_data)

        return results[0]



import librosa
import numpy as np

from nlpaug.model.audio import Audio

"""
    Reference: https://www.kaggle.com/CVxTz/audio-data-augmentation
    A wrapper of librosa.effects.time_stretch
"""

# TODO: Validation


class Speed(Audio):
    """
    Adjusting speed

    :param speed_range: Factor for time stretch. Audio will be slowing down if value is between 0 and 1.
        Audio will be speed up if value is larger than 1.
    """
    def __init__(self, speed_range):
        super(Speed, self).__init__()

        # if speed_factor < 0:
        #     raise ValueError(
        #         'speed_factor should be positive number while {} is passed.'.format(speed_factor))
        self.speed_range = speed_range

    def manipulate(self, data):
        speeds = [round(i, 1) for i in np.arange(self.speed_range[0], self.speed_range[1], 0.1)]
        speed = speeds[np.random.randint(len(speeds))]

        return librosa.effects.time_stretch(data, speed)



import numpy as np

from nlpaug.model.audio import Audio

"""
    Reference: https://www.kaggle.com/CVxTz/audio-data-augmentation
"""


class Shift(Audio):
    def __init__(self, sampling_rate, shift_max=2, shift_direction='both'):
        """
         :param sampling_rate: SR of audio
         :param shift_max: Max shifting in second
         :param shift_direction: Shifting segment to left, right or one of them
         """

        super(Shift, self).__init__()

        self.sampling_rate = sampling_rate
        self.shift_max = shift_max

        if shift_direction in ['left', 'right', 'both']:
            self.shift_direction = shift_direction
        else:
            raise ValueError(
                'shift_direction should be either left, right or both while {} is passed.'.format(shift_direction))

    def manipulate(self, data):
        shift = np.random.randint(self.sampling_rate * self.shift_max)
        if self.shift_direction == 'right':
            shift = -shift
        elif self.shift_direction == 'both':
            direction = np.random.randint(0, 2)
            if direction == 1:
                shift = -shift

        augmented_data = np.roll(data, shift)

        # Set to silence for heading/ tailing
        if shift > 0:
            augmented_data[:shift] = 0
        else:
            augmented_data[shift:] = 0
        return augmented_data



import numpy as np

def standard_norm(data):
    means = data.mean(axis =1)
    stds = data.std(axis= 1, ddof=1)
    data = (data - means[:, np.newaxis]) / stds[:, np.newaxis]
    return np.nan_to_num(data)

def l1_norm(data):
    _norm = np.array([x.sum(axis=0) for x in data])
    data = data/_norm[:, np.newaxis]
    return np.nan_to_num(data)

def l2_norm(data):
    _norm = np.array([np.sqrt((x*x).sum(axis=0)) for x in data])
    data = data/_norm[:, np.newaxis]
    return np.nan_to_num(data)



"""
    Augmenter that apply stopwords removal to textual input.
"""

from nlpaug.augmenter.word import WordAugmenter
from nlpaug.util import Action, WarningException, WarningName, WarningCode, WarningMessage
from nlpaug.util.decorator.deprecation import deprecated


@deprecated(deprecate_from='0.0.7', deprecate_to='0.0.9', msg="Use RandomWordAug from 0.0.7 version")
class StopWordsAug(WordAugmenter):
    """
    Augmenter that leverage pre-defined spelling mistake dictionary to simulate spelling mistake.

    :param list stopwords: List of stopword will be removed.
    :param int aug_min: Minimum number of word will be augmented.
    :param float aug_p: Percentage of word will be augmented.
    :param func tokenizer: Customize tokenization process
    :param func reverse_tokenizer: Customize reverse of tokenization process
    :param str name: Name of this augmenter

    >>> import nlpaug.augmenter.word as naw
    >>> aug = naw.StopWordsAug(stopwords=['a', 'an', 'the'])
    """

    def __init__(self, stopwords, name='StopWords_Aug', aug_min=1, aug_p=0.3,
                 tokenizer=None, reverse_tokenizer=None, case_sensitive=False, verbose=0):

        if not case_sensitive:
            stopwords = [t.lower() for t in stopwords]

        super().__init__(
            action=Action.DELETE, name=name, aug_p=aug_p, aug_min=aug_min, stopwords=stopwords,
            tokenizer=tokenizer, reverse_tokenizer=reverse_tokenizer, verbose=verbose)

        self.case_sensitive = case_sensitive

    def _get_aug_idxes(self, tokens):
        aug_cnt = self.generate_aug_cnt(len(tokens))
        word_idxes = []
        for i, t in enumerate(tokens):
            token = t
            if not self.case_sensitive:
                token = token.lower()
            if token in self.stopwords:
                word_idxes.append(i)

        word_idxes = self.skip_aug(word_idxes, tokens)
        if len(word_idxes) == 0:
            if self.verbose > 0:
                exception = WarningException(name=WarningName.OUT_OF_VOCABULARY,
                                             code=WarningCode.WARNING_CODE_002, msg=WarningMessage.NO_WORD)
                exception.output()
            return None
        if len(word_idxes) < aug_cnt:
            aug_cnt = len(word_idxes)
        aug_idexes = self.sample(word_idxes, aug_cnt)
        return aug_idexes

    def delete(self, data):
        tokens = self.tokenizer(data)
        results = tokens.copy()

        aug_idxes = self._get_aug_idxes(tokens)
        if aug_idxes is None:
            return data

        aug_idxes.sort(reverse=True)

        for aug_idx in aug_idxes:
            del results[aug_idx]

        if len(results) > 0 and len(results[0]) > 0:
            results[0] = self.align_capitalization(tokens[0], results[0])

        return self.reverse_tokenizer(results)



from nlpaug.util import Method
from nlpaug import Augmenter


class SpectrogramAugmenter(Augmenter):
    def __init__(self, action, name='Spectrogram_Aug', aug_min=1, aug_p=0.3, verbose=0):
        super(SpectrogramAugmenter, self).__init__(
            name=name, method=Method.SPECTROGRAM, action=action, aug_min=aug_min, verbose=verbose)
        self.aug_p = aug_p




import functools
import warnings


def deprecated(deprecate_from, deprecate_to, msg):
    def decorator(obj):
        if isinstance(obj, type):
            return _decorate_class(obj, deprecate_from, deprecate_to, msg)
        # # TODO:
        # elif isinstance(obj, property):
        #     return _decorate_prop(obj, msg)
        else:
            return _decorate_func(obj, deprecate_from, deprecate_to, msg)
    return decorator


def _decorate_class(cls, deprecate_from, deprecate_to, msg):
    msg_template = 'Class {name} is deprecated from {deprecate_from} version.'
    msg_template += ' It will be removed from {deprecate_to} version. {msg}'

    @functools.wraps(cls)
    def wrapped(*args, **kwargs):
        warnings.simplefilter('always', DeprecationWarning)
        warnings.warn(
            msg_template.format(
                name=cls.__name__, deprecate_from=deprecate_from, deprecate_to=deprecate_to, msg=msg),
            category=DeprecationWarning
        )
        warnings.simplefilter('default', DeprecationWarning)
        return cls(*args, **kwargs)

    return wrapped


def _decorate_func(func, deprecate_from, deprecate_to, msg):
    msg_template = 'Function {name} is deprecated from {deprecate_from} version.'
    msg_template += ' It will be removed from {deprecate_to} version. {msg}'

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        warnings.simplefilter('always', DeprecationWarning)
        warnings.warn(
            msg_template.format(
                name=func.__name__, deprecate_from=deprecate_from, deprecate_to=deprecate_to, msg=msg),
            category=DeprecationWarning
        )
        warnings.simplefilter('default', DeprecationWarning)
        return func(*args, **kwargs)

    return wrapped


def _decorate_prop(prop, msg):
    @functools.wraps(prop)
    @property
    def wrapped(*args, **kwargs):
        msg_template = 'Property {name} is deprecated. {msg}'
        warnings.simplefilter('always', DeprecationWarning)
        warnings.warn(
            msg_template.format(name=prop.__name__, msg=msg), category=DeprecationWarning
        )
        warnings.simplefilter('default', DeprecationWarning)
        return prop.fget(*args, **kwargs)

    return wrapped



"""
    Augmenter that apply semantic meaning based to textual input.
"""

import nltk
from nltk.corpus import wordnet

from nlpaug.augmenter.word import WordAugmenter
from nlpaug.util import Action, PartOfSpeech, WarningException, WarningName, WarningCode, WarningMessage


class WordNetAug(WordAugmenter):
    """
    Augmenter that leverage semantic meaning to substitute word.

    :param str lang: Language of your text. Default value is 'eng'.
    :param int aug_min: Minimum number of word will be augmented.
    :param float aug_p: Percentage of word will be augmented.
    :param list stopwords: List of words which will be skipped from augment operation.
    :param func tokenizer: Customize tokenization process
    :param func reverse_tokenizer: Customize reverse of tokenization process
    :param str name: Name of this augmenter

    >>> import nlpaug.augmenter.word as naw
    >>> aug = naw.WordNetAug()
    """

    def __init__(self, name='WordNet_Aug', aug_min=1, aug_p=0.3, lang='eng', stopwords=None,
                 tokenizer=None, reverse_tokenizer=None, verbose=0):
        super().__init__(
            action=Action.SUBSTITUTE, name=name, aug_p=aug_p, aug_min=aug_min, stopwords=stopwords,
            tokenizer=tokenizer, reverse_tokenizer=reverse_tokenizer, verbose=verbose)

        self.model = self.get_model()
        self.lang = lang

        ### TODO: antonym: https://arxiv.org/pdf/1809.02079.pdf
        self.synonyms = True

    def skip_aug(self, token_idxes, tokens):
        results = []
        for token_idx in token_idxes:
            # Some word does not come with synonym. It will be excluded in lucky draw.
            if tokens[token_idx][1] not in ['DT']:
                results.append(token_idx)

        return results

    def _get_aug_idxes(self, tokens):
        aug_cnt = self.generate_aug_cnt(len(tokens))
        word_idxes = [i for i, t in enumerate(tokens) if self.stopwords is None or t[0] not in self.stopwords]
        word_idxes = self.skip_aug(word_idxes, tokens)
        if len(word_idxes) == 0:
            if self.verbose > 0:
                exception = WarningException(name=WarningName.OUT_OF_VOCABULARY,
                                             code=WarningCode.WARNING_CODE_002, msg=WarningMessage.NO_WORD)
                exception.output()
            return None
        if len(word_idxes) < aug_cnt:
            aug_cnt = len(word_idxes)
        aug_idexes = self.sample(word_idxes, aug_cnt)
        return aug_idexes

    def substitute(self, data):
        results = []

        tokens = self.tokenizer(data)

        pos = nltk.pos_tag(tokens)

        aug_idxes = self._get_aug_idxes(pos)
        if aug_idxes is None:
            return data

        for i, token in enumerate(tokens):
            # Skip if no augment for word
            if i not in aug_idxes:
                results.append(token)
                continue

            word_poses = PartOfSpeech.pos2wn(pos[i][1])
            synets = []
            if word_poses is None or len(word_poses) == 0:
                # Use every possible words as the mapping does not defined correctly
                synets.extend(self.model.synsets(pos[i][0], lang=self.lang))
            else:
                for word_pos in word_poses:
                    synets.extend(self.model.synsets(pos[i][0], pos=word_pos, lang=self.lang))

            augmented_data = []
            for synet in synets:
                candidates = []
                for lema in synet.lemmas():
                    if self.synonyms:
                        candidates.append(lema.name())
                    else:
                        if lema.antonyms():
                            candidates.append(lema.antonyms()[0].name())

                for candidate in candidates:
                    if candidate.lower() != token.lower():
                        augmented_data.append(candidate)

            if len(augmented_data) == 0:
                results.append(token)
            else:
                candidate = self.sample(augmented_data, 1)[0]
                candidate = candidate.replace("_", " ").replace("-", " ").lower()
                results.append(self.align_capitalization(token, candidate))

        return self.reverse_tokenizer(results)

    @classmethod
    def get_model(cls):
        try:
            # Check whether wordnet package is downloaded
            wordnet.synsets('computer')
        except Exception:
            nltk.download('wordnet')

        try:
            # Check whether POS package is downloaded
            nltk.pos_tag('computer')
        except Exception:
            nltk.download('averaged_perceptron_tagger')

        return wordnet



import numpy as np

from nlpaug.model.audio import Audio


class Noise(Audio):
    """
        Reference: https://www.kaggle.com/CVxTz/audio-data-augmentation
    """

    def __init__(self, noise_factor):
        """

        :param noise_factor: Strength of noise
        """
        super(Noise, self).__init__()

        self.noise_factor = noise_factor

    def manipulate(self, data):
        noise = np.random.randn(len(data))
        augmented_data = data + self.noise_factor * noise
        # Cast back to same data type
        augmented_data = augmented_data.astype(type(data[0]))
        return augmented_data



from nlpaug.util import Method
from nlpaug import Augmenter
from nlpaug.util import WarningException, WarningName, WarningCode, WarningMessage


class WordAugmenter(Augmenter):
    def __init__(self, action, name='Word_Aug', aug_min=1, aug_p=0.3, stopwords=None,
                 tokenizer=None, reverse_tokenizer=None, verbose=0):
        super().__init__(
            name=name, method=Method.WORD, action=action, aug_min=aug_min, verbose=verbose)
        self.aug_p = aug_p
        self.tokenizer = tokenizer or self._tokenizer
        self.reverse_tokenizer = reverse_tokenizer or self._reverse_tokenizer
        self.stopwords = stopwords

    @classmethod
    def _tokenizer(cls, text):
        return text.split(' ')

    @classmethod
    def _reverse_tokenizer(cls, tokens):
        return ' '.join(tokens)

    def skip_aug(self, token_idxes, tokens):
        return token_idxes

    @classmethod
    def align_capitalization(cls, src_token, dest_token):
        # For whole word is upper case
        if src_token[0].isupper() and len(src_token) > 1 and src_token[1].isupper():
            return dest_token.upper()
        # For capitalize word
        elif src_token[0].isupper():
            return dest_token.capitalize()
        else:
            return dest_token

    def _get_aug_idxes(self, tokens):
        aug_cnt = self.generate_aug_cnt(len(tokens))
        word_idxes = [i for i, t in enumerate(tokens) if self.stopwords is None or t not in self.stopwords]
        word_idxes = self.skip_aug(word_idxes, tokens)
        if len(word_idxes) == 0:
            if self.verbose > 0:
                exception = WarningException(name=WarningName.OUT_OF_VOCABULARY,
                                             code=WarningCode.WARNING_CODE_002, msg=WarningMessage.NO_WORD)
                exception.output()
            return None
        if len(word_idxes) < aug_cnt:
            aug_cnt = len(word_idxes)
        aug_idexes = self.sample(word_idxes, aug_cnt)
        return aug_idexes

    def _get_random_aug_idxes(self, tokens):
        aug_cnt = self.generate_aug_cnt(len(tokens))
        word_idxes = [i for i in range(len(tokens))]
        if len(word_idxes) < aug_cnt:
            aug_cnt = len(word_idxes)

        aug_idxes = self.sample(word_idxes, aug_cnt)

        return aug_idxes



from nlpaug.util.action import *
from nlpaug.util.operation import *
from nlpaug.util.method import *
from nlpaug.util.exception import *
from nlpaug.util.math import *
from nlpaug.util.text import *

from nlpaug.util.part_of_speech import *

from nlpaug.util.file import *
from nlpaug.util.decorator import *


import librosa


class LoadUtil:
    @staticmethod
    def load_mel_spectrogram(file_path, n_mels=128, fmax=8000):
        audio, sampling_rate = librosa.load(file_path)
        return librosa.feature.melspectrogram(
            y=audio, sr=sampling_rate, n_mels=n_mels, fmax=fmax)


import unittest


if __name__ == '__main__':
    test_dirs = [
        'test/augmenter/char/',
        'test/augmenter/word/',
        'test/augmenter/audio/',
        'test/augmenter/spectrogram/',
        'test/flow/'
    ]

    runner = unittest.TextTestRunner()

    for test_dir in test_dirs:
       loader = unittest.TestLoader()
       suite = loader.discover(test_dir)
       runner.run(suite)

    # suite = unittest.TestLoader().loadTestsFromName('augmenter.word.test_wordembs')
    # runner.run(suite)



import os, urllib, zipfile, requests


class DownloadUtil:
    @staticmethod
    def download_word2vec(dest_dir):
        DownloadUtil.download_from_google_drive(
            _id='0B7XkCwpI5KDYNlNUTTlSS21pQmM', dest_dir=dest_dir, dest_file='GoogleNews-vectors-negative300.zip'
        )

    @staticmethod
    def download_glove(model_name, dest_dir):
        url = ''
        if model_name == 'glove.6B':
            url = 'http://nlp.stanford.edu/data/glove.6B.zip'
        elif model_name == 'glove.42B.300d':
            url = 'http://nlp.stanford.edu/data/glove.42B.300d.zip'
        elif model_name == 'glove.840B.300d':
            url = 'http://nlp.stanford.edu/data/glove.840B.300d.zip'
        elif model_name == 'glove.twitter.27B':
            url = 'http://nlp.stanford.edu/data/glove.twitter.27B.zip',
        else:
            possible_values = ['glove.6B', 'glove.42B.300d', 'glove.840B.300d', 'glove.twitter.27B']
            raise ValueError('Unknown model_name. Possible values are {}'.format(possible_values))

        file_path = DownloadUtil.download(url, dest_dir=dest_dir)
        DownloadUtil.unzip(file_path)

    @staticmethod
    def download_fasttext(model_name, dest_dir):
        url = ''
        if model_name == 'wiki-news-300d-1M':
            url = 'https://dl.fbaipublicfiles.com/fasttext/vectors-english/wiki-news-300d-1M.vec.zip'
        # elif model_name == 'wiki-news-300d-1M-subword':
        #     url = 'https://dl.fbaipublicfiles.com/fasttext/vectors-english/wiki-news-300d-1M-subword.vec.zip'
        elif model_name == 'crawl-300d-2M':
            url = 'https://dl.fbaipublicfiles.com/fasttext/vectors-english/crawl-300d-2M.vec.zip'
        # elif model_name == 'crawl-300d-2M-subword':
        #     url = 'https://dl.fbaipublicfiles.com/fasttext/vectors-english/crawl-300d-2M-subword.zip'
        else:
            possible_values = ['wiki-news-300d-1M', 'crawl-300d-2M']
            raise ValueError('Unknown model_name. Possible values are {}'.format(possible_values))

        file_path = DownloadUtil.download(url, dest_dir=dest_dir)
        DownloadUtil.unzip(file_path)

    @staticmethod
    def download(src, dest_dir, dest_file=None):
        if not os.path.exists(dest_dir):
            os.makedirs(dest_dir)

        if dest_file is None:
            dest_file = os.path.basename(src)

        if not os.path.exists(dest_dir + dest_file):
            req = urllib.request.Request(src)
            file = urllib.request.urlopen(req)
            with open(dest_dir + dest_file, 'wb') as output:
                output.write(file.read())
        return dest_dir + dest_file

    @staticmethod
    def unzip(file_path):
        dest_dir = os.path.dirname(file_path)
        with zipfile.ZipFile(file_path, "r") as zip_ref:
            zip_ref.extractall(dest_dir)

    @staticmethod
    def download_from_google_drive(_id, dest_dir, dest_file):
        url = "https://docs.google.com/uc?export=download"
        def get_confirm_token(response):
            for key, value in response.cookies.items():
                if key.startswith('download_warning'):
                    return value

            return None

        def save_response_content(response, destination):
            CHUNK_SIZE = 32768

            with open(destination, "wb") as f:
                for chunk in response.iter_content(CHUNK_SIZE):
                    if chunk:  # filter out keep-alive new chunks
                        f.write(chunk)

        session = requests.Session()

        response = session.get(url, params={'id': _id}, stream=True)
        token = get_confirm_token(response)

        if token:
            params = {'id': _id, 'confirm': token}
            response = session.get(url, params=params, stream=True)

        save_response_content(response, os.path.join(dest_dir, dest_file))



"""
    Augmenter that apply typo error simulation to textual input.
"""

import re

from nlpaug.augmenter.char import CharAugmenter
from nlpaug.util import Action, Method
from nlpaug.util.decorator.deprecation import deprecated


@deprecated(deprecate_from='0.0.7', deprecate_to='0.0.9', msg="Use KeyboardAug from 0.0.7 version")
class QwertyAug(CharAugmenter):
    """
    Augmenter that simulate typo error by random values. For example, people may type i as o incorrectly.\
        One keyboard distance is leveraged to replace character by possible keyboard error.

    :param int aug_min: Minimum number of character will be augmented.
    :param float aug_char_p: Percentage of character (per token) will be augmented.
    :param float aug_word_p: Percentage of word will be augmented.
    :param list stopwords: List of words which will be skipped from augment operation.
    :param func tokenizer: Customize tokenization process
    :param func reverse_tokenizer: Customize reverse of tokenization process
    :param str name: Name of this augmenter

    >>> import nlpaug.augmenter.char as nac
    >>> aug = nac.QwertyAug()
    """

    def __init__(self, name='Qwerty_Aug', aug_min=1, aug_char_p=0.3, aug_word_p=0.3, stopwords=None,
                 tokenizer=None, reverse_tokenizer=None, verbose=0):
        super().__init__(
            action=Action.SUBSTITUTE, name=name, aug_char_p=aug_char_p, aug_word_p=aug_word_p, aug_min=aug_min,
            tokenizer=tokenizer, reverse_tokenizer=reverse_tokenizer, stopwords=stopwords, verbose=verbose)

        self.model = self.get_model()

    def skip_aug(self, token_idxes, tokens):
        results = []
        for token_idx in token_idxes:
            # Some word does not come with vector. It will be excluded in lucky draw.
            char = tokens[token_idx]
            if char in self.model and len(self.model[char]) > 1:
                results.append(token_idx)

        return results

    def substitute(self, data):
        results = []
        tokens = self.tokenizer(data)
        aug_word_idxes = self._get_aug_idxes(tokens, self.aug_word_p, Method.WORD)

        for token_i, token in enumerate(tokens):
            if token_i not in aug_word_idxes:
                results.append(token)
                continue

            result = ''
            chars = self.token2char(token)
            aug_char_idxes = self._get_aug_idxes(chars, self.aug_char_p, Method.CHAR)
            if aug_char_idxes is None:
                results.append(token)
                continue

            for char_i, char in enumerate(chars):
                if char_i not in aug_char_idxes:
                    result += char
                    continue

                result += self.sample(self.model[chars[char_i]], 1)[0]

            results.append(result)

        return self.reverse_tokenizer(results)

    # TUNEME: pre calcualte it
    def get_model(self, include_spec=True, case_sensitive=True):
        mapping = {
            '1': ['!', '2', '@', 'q', 'w'],
            '2': ['@', '1', '!', '3', '#', 'q', 'w', 'e'],
            '3': ['#', '2', '@', '4', '$', 'w', 'e'],
            '4': ['$', '3', '#', '5', '%', 'e', 'r'],
            '5': ['%', '4', '$', '6', '^', 'r', 't', 'y'],
            '6': ['^', '5', '%', '7', '&', 't', 'y', 'u'],
            '7': ['&', '6', '^', '8', '*', 'y', 'u', 'i'],
            '8': ['*', '7', '&', '9', '(', 'u', 'i', 'o'],
            '9': ['(', '8', '*', '0', ')', 'i', 'o', 'p'],

            'q': ['1', '!', '2', '@', 'w', 'a', 's'],
            'w': ['1', '!', '2', '@', '3', '#', 'q', 'e', 'a', 's', 'd'],
            'e': ['2', '@', '3', '#', '4', '$', 'w', 'r', 's', 'd', 'f'],
            'r': ['3', '#', '4', '$', '5', '%', 'e', 't', 'd', 'f', 'g'],
            't': ['4', '$', '5', '%', '6', '^', 'r', 'y', 'f', 'g', 'h'],
            'y': ['5', '%', '6', '^', '7', '&', 't', 'u', 'g', 'h', 'j'],
            'u': ['6', '^', '7', '&', '8', '*',' t', 'i', 'h', 'j', 'k'],
            'i': ['7', '&', '8', '*', '9', '(', 'u', 'o', 'j', 'k', 'l'],
            'o': ['8', '*', '9', '(', '0', ')', 'i', 'p', 'k', 'l'],
            'p': ['9', '(', '0', ')', 'o', 'l'],

            'a': ['q', 'w', 'a', 's', 'z', 'x'],
            's': ['q', 'w', 'e', 'a', 'd', 'z', 'x', 'c'],
            'd': ['w', 'e', 'r', 's', 'f', 'x', 'c', 'v'],
            'f': ['e', 'r', 't', 'd', 'g', 'c', 'v', 'b'],
            'g': ['r', 't', 'y', 'f', 'h', 'v', 'b', 'n'],
            'h': ['t', 'y', 'u', 'g', 'j', 'b', 'n', 'm'],
            'j': ['y', 'u', 'i', 'h', 'k', 'n', 'm', ',', '<'],
            'k': ['u', 'i', 'o', 'j', 'l', 'm', ',', '<', '.', '>'],
            'l': ['i', 'o', 'p', 'k', ';', ':', ',', '<', '.', '>', '/', '?'],

            'z': ['a', 's', 'x'],
            'x': ['a', 's', 'd', 'z', 'c'],
            'c': ['s', 'd', 'f', 'x', 'v'],
            'v': ['d', 'f', 'g', 'c', 'b'],
            'b': ['f', 'g', 'h', 'v', 'n'],
            'n': ['g', 'h', 'j', 'b', 'm'],
            'm': ['h', 'j', 'k', 'n', ',', '<']
        }

        result = {}

        for k in mapping:
            result[k] = []
            for v in mapping[k]:
                if not include_spec and not re.match("^[a-zA-Z0-9]*$", v):
                    continue

                result[k].append(v)

                if case_sensitive:
                    result[k].append(v.upper())

            result[k] = list(set(result[k]))

            if case_sensitive and re.match("^[a-z]*$", k):
                result[k.upper()] = []

                for v in mapping[k]:
                    if not include_spec and not re.match("^[a-zA-Z0-9]*$", v):
                        continue

                    k = k.upper()

                    result[k].append(v)

                    if case_sensitive:
                        result[k].append(v.upper())

                result[k] = list(set(result[k]))

        return result


import matplotlib.pyplot as plt
import librosa.display


class VisualWave:
    @staticmethod
    def visual(title, audio, sample_rate):
        plt.figure(figsize=(8, 4))
        librosa.display.waveplot(audio, sr=sample_rate)
        plt.title(title)
        plt.tight_layout()
        plt.show()



from nlpaug.util.exception.exception_info import ExceptionInfo, ExceptionType


class WarningException(ExceptionInfo):
    def __init__(self, name, code, msg):
        super(WarningException, self).__init__(name=name, exp_type=ExceptionType.WARNING, code=code, msg=msg)


class WarningName:
    INPUT_VALIDATION_WARNING = 'Input validation issue'
    OUT_OF_VOCABULARY = 'Out of vocabulary issue'


class WarningCode:
    WARNING_CODE_001 = 'W001'
    WARNING_CODE_002 = 'W002'


class WarningMessage:
    LENGTH_IS_ZERO = 'Length of input is 0'
    NO_WORD = 'No other word except stop words and OOV. Returning input data without augmentation'



"""
    Flow that apply augmentation randomly.
"""

from nlpaug.util import Action
from nlpaug.flow import Pipeline


class Sometimes(Pipeline):
    """
    Flow that apply augmenters randomly.

    :param list flow: list of flow or augmenter
    :param str name: Name of this augmenter

    >>> import nlpaug.flow as naf
    >>> import nlpaug.augmenter.char as nac
    >>> import nlpaug.augmenter.word as naw
    >>> flow = naf.Sometimes([nac.RandomCharAug(), naw.RandomWordAug()])
    """

    # TODO: Using epoch to implement 1-to-many
    def __init__(self, flow=None, name='Sometimes_Pipeline', pipeline_p=0.2, aug_p=1, verbose=0):
        Pipeline.__init__(self, name=name, action=Action.SOMETIMES,
                          flow=flow, epoch=1, aug_min=-1, aug_p=aug_p, verbose=verbose)

        self.pipeline_p = pipeline_p

    def augment(self, data):
        """
        :param data: Data for augmentation
        :return: Augmented data

        >>> augmented_data = flow.augment(data)
        """
        results = []

        for _ in range(self.epoch):
            augmented_data = data[:]
            for aug in self:
                if self.pipeline_p < self.prob():
                    continue

                augmented_data = aug.augment(augmented_data)

            results.append(augmented_data)

        return results[0]



import random
import numpy as np

from nlpaug.util import Action, Method, Operation, WarningException, WarningName, WarningCode, WarningMessage


class Augmenter:
    def __init__(self, name, method, action, aug_min, aug_p=0.1, verbose=0):
        self.name = name
        self.action = action
        self.method = method
        self.aug_min = aug_min
        self.aug_p = aug_p
        self.verbose = verbose
        
        self.augments = []
        
        self._validate_augmenter(method, action)
        
    def _init_aug_idxes(self, aug_p):
        self.aug_per_idxes = []
        if isinstance(aug_p, list):
            self.aug_num_mode = Operation.RANGE_LIST_PERCENTAGE
            self.aug_pers = aug_p
        elif isinstance(aug_p, tuple):
            self.aug_num_mode = Operation.RANGE_TUPLE_PERCENTAGE
            self.aug_pers = [i/10 for i in range(int(aug_p[0]*10), int(aug_p[1]*10)+1)]
        elif isinstance(aug_p, float):
            self.aug_num_mode = Operation.EXACT_PERCENTAGE
            self.aug_pers = [aug_p]
        else:
            raise ValueError(
                'aug_per should be list, tuple of float while {} is passed.'.format(type(self.aug_p)))

    @classmethod
    def _validate_augmenter(cls, method, action):
        if method not in Method.getall():
            raise ValueError(
                'Method must be one of {} while {} is passed'.format(Method.getall(), method))

        if action not in Action.getall():
            raise ValueError(
                'Action must be one of {} while {} is passed'.format(Action.getall(), action))
                
    def augment(self, data):
        """
        :param data: Data for augmentation
        :return: Augmented data

        >>> augmented_data = aug.augment(data)

        """
        exceptions = self._validate_augment(data)
        # TODO: Handle multiple exceptions
        for exception in exceptions:
            if isinstance(exception, WarningException):
                if self.verbose > 0:
                    exception.output()

                # Return empty value per data type
                if isinstance(data, str):
                    return ''
                elif isinstance(data, list):
                    return []
                elif isinstance(data, np.ndarray):
                    return np.array([])

                return None

        if self.action == Action.INSERT:
            return self.insert(data)
        elif self.action == Action.SUBSTITUTE:
            return self.substitute(data)
        elif self.action == Action.SWAP:
            return self.swap(data)
        elif self.action == Action.DELETE:
            return self.delete(data)

    @classmethod
    def _validate_augment(cls, data):
        if data is None or len(data) == 0:
            return [WarningException(name=WarningName.INPUT_VALIDATION_WARNING,
                                     code=WarningCode.WARNING_CODE_001, msg=WarningMessage.LENGTH_IS_ZERO)]

        return []

    def insert(self, data):
        raise NotImplementedError()

    def substitute(self, data):
        raise NotImplementedError()

    def swap(self, data):
        raise NotImplementedError()

    def delete(self, data):
        raise NotImplementedError()
        
    def tokenizer(self, tokens):
        raise NotImplementedError()

    def evaluate(self):
        raise NotImplementedError()

    @classmethod
    def prob(cls):
        return random.random()

    @classmethod
    def sample(cls, x, num):
        return random.sample(x, num)

    def generate_aug_cnt(self, size, aug_p=None):
        if aug_p is not None:
            percent = aug_p
        elif self.aug_p is not None:
            percent = self.aug_p
        else:
            percent = 0.3
        cnt = int(percent * size)
        return cnt if cnt > self.aug_min else self.aug_min

    def generate_aug_idxes(self, inputs):
        aug_cnt = self.generate_aug_cnt(len(inputs))
        token_idxes = [i for i, _ in enumerate(inputs)]
        aug_idxes = self.sample(token_idxes, aug_cnt)
        return aug_idxes

    def __str__(self):
        return 'Name:{}, Action:{}, Method:{}'.format(self.name, self.action, self.method)


"""
    Augmenter that apply fasttext's based operation to textual input.
"""

from nlpaug.augmenter.word import WordEmbsAugmenter
from nlpaug.util import Action
import nlpaug.model.word_embs as nmw
from nlpaug.util.decorator.deprecation import deprecated

FASTTEXT_MODEL = {}


def init_fasttext_model(model_path, force_reload=False):
    """
        Load model once at runtime
    """
    global FASTTEXT_MODEL
    if FASTTEXT_MODEL and not force_reload:
        return FASTTEXT_MODEL

    fasttext = nmw.Fasttext()
    fasttext.read(model_path)
    FASTTEXT_MODEL = fasttext

    return FASTTEXT_MODEL


@deprecated(deprecate_from='0.0.7', deprecate_to='0.0.9', msg="Use WordEmbsAug from 0.0.7 version")
class FasttextAug(WordEmbsAugmenter):
    """
    Augmenter that leverage fasttext's embeddings to find top n similar word for augmentation.

    :param str model_path: Downloaded model directory. Either model_path or model is must be provided
    :param obj model: Pre-loaded model
    :param str action: Either 'insert or 'substitute'. If value is 'insert', a new word will be injected to random
        position according to word embeddings calculation. If value is 'substitute', word will be replaced according
        to word embeddings calculation
    :param int aug_min: Minimum number of word will be augmented.
    :param float aug_p: Percentage of word will be augmented.
    :param int aug_n: Top n similar word for lucky draw
    :param list stopwords: List of words which will be skipped from augment operation.
    :param func tokenizer: Customize tokenization process
    :param func reverse_tokenizer: Customize reverse of tokenization process
    :param bool force_reload: If True, model will be loaded every time while it takes longer time for initialization.
    :param str name: Name of this augmenter

    >>> import nlpaug.augmenter.word as naw
    >>> aug = naw.FasttextAug()
    """

    def __init__(self, model_path='.', model=None, action=Action.SUBSTITUTE,
                 name='Fasttext_Aug', aug_min=1, aug_p=0.3, aug_n=5, stopwords=None,
                 tokenizer=None, reverse_tokenizer=None, force_reload=False,
                 verbose=0):
        super().__init__(
            model_path=model_path, aug_n=aug_n,
            action=action, name=name, aug_p=aug_p, aug_min=aug_min, stopwords=stopwords,
            tokenizer=tokenizer, reverse_tokenizer=reverse_tokenizer, verbose=verbose)

        if model is None:
            self.model = self.get_model(force_reload=force_reload)
        else:
            self.model = model

    def get_model(self, force_reload=False):
        return init_fasttext_model(self.model_path, force_reload)



import numpy as np

from nlpaug.model.word_embs import WordEmbeddings


class Word2vec(WordEmbeddings):
    def __init__(self, cache=True, skip_check=False):
        super().__init__(cache, skip_check)

    def read(self, file_path, max_num_vector=None):
        with open(file_path, 'rb') as f:
            header = f.readline()
            self.vocab_size, self.emb_size = map(int, header.split())
            if max_num_vector is not None:
                self.vocab_size = min(max_num_vector, self.vocab_size)

            self.vectors = np.zeros((self.vocab_size, self.emb_size), dtype=np.float32)
            binary_len = np.dtype(np.float32).itemsize * self.emb_size

            for _ in range(self.vocab_size):
                word = []
                while True:
                    ch = f.read(1)
                    if ch == b' ':
                        word = ''.join(word)
                        break
                    if ch != '\n':
                        word.append(ch.decode('cp437'))
                values = np.fromstring(f.read(binary_len), dtype=np.float32)

                # word = " ".join(tokens[0:(len(tokens) - self.emb_size):])
                # values = np.array([float(val) for val in tokens[(self.emb_size * -1):]])

                self.vectors[len(self.i2w)] = values
                self.i2w[len(self.i2w)] = word
                self.w2i[word] = len(self.w2i)
                self.w2v[word] = values

        self.vectors = np.asarray(self.vectors)
        if not self.skip_check:
            if len(self.vectors) != len(self.i2w):
                raise AssertionError('Vector Size:{}, Index2Word Size:{}'.format(len(self.vectors), len(self.i2w)))
            if len(self.i2w) != len(self.w2i):
                raise AssertionError('Index2Word Size:{}, Word2Index Size:{}'.format(len(self.i2w), len(self.w2i)))
            if len(self.w2i) != len(self.w2v):
                raise AssertionError('Word2Index Size:{}, Word2Vector Size:{}'.format(len(self.w2i), len(self.w2v)))

        self.normalized_vectors = self._normalize(self.vectors)

        if self.cache:
            self.vocab = [word for word in self.w2v]




import matplotlib.pyplot as plt
import librosa.display
import numpy as np


class VisualSpectrogram:
    @staticmethod
    def visual(title, spectrogram):
        plt.figure(figsize=(8, 4))
        librosa.display.specshow(
            librosa.power_to_db(spectrogram, ref=np.max), y_axis='mel', fmax=8000, x_axis='time')
        plt.colorbar(format='%+10.0f dB')
        plt.title(title)
        plt.tight_layout()
        plt.show()



import numpy as np

from nlpaug.model.audio import Audio


class Loudness(Audio):
    def __init__(self, loudness_factor=(0.5, 2)):
        """

        :param loudness_factor: Input data volume will be increased (decreased). Augmented value will be picked
            within the range of this tuple value. If volume will be reduced if value is between 0 and 1.
        """
        super().__init__()
        self.loudness_factor = loudness_factor

    def manipulate(self, data):
        loud = np.random.uniform(self.loudness_factor[0], self.loudness_factor[1])
        augmented_data = data * loud
        return augmented_data



import numpy as np

from nlpaug.model.audio import Audio


class Mask(Audio):
    def __init__(self, sampling_rate, mask_range=(0.2, 0.8), mask_factor=2, mask_with_noise=True):
        """

        :param sampling_rate: sampling rate of input audio
        :param mask_range: Range of applying mask operation. Default value is (0.2, 0.8)
            It means that first 20% and last 20% of data will not be excluded from augment operation. Augment operation
            will be applied to clip of rest of 60% time.
        :param mask_factor: duration of masking period (in second)
        :param mask_with_noise: If it is True, targeting area will be replaced by noise. Otherwise, it will be
            replaced by 0.
        """
        super().__init__()
        self.sampling_rate = sampling_rate
        self.mask_range = mask_range
        self.mask_factor = mask_factor
        self.mask_with_noise = mask_with_noise

    def manipulate(self, data):
        valid_region = (int(len(data) * self.mask_range[0]), int(len(data) * self.mask_range[1]))

        start_timeframe = np.random.randint(valid_region[0], valid_region[1])
        end_timeframe = start_timeframe + self.sampling_rate * self.mask_factor

        # Mask region is larger than valid region
        if end_timeframe > valid_region[1]:
            end_timeframe = valid_region[1]

        masked_value = None
        if self.mask_with_noise:
            masked_value = np.random.randn(end_timeframe - start_timeframe)
        else:
            masked_value = [0 for _ in range(end_timeframe-start_timeframe)]

        augmented_data = data.copy()
        augmented_data[start_timeframe:end_timeframe] = masked_value

        return augmented_data



"""
    Augmenter that apply noise injection operation to audio.
"""

from nlpaug.augmenter.audio import AudioAugmenter
from nlpaug.util import Action
import nlpaug.model.audio as nma


# TODO: Apply noise injection to segment of audio only

class NoiseAug(AudioAugmenter):
    """
    Augmenter that crop segment of audio by random values between crop_range variable.

    :param int noise_factor: Strength of noise
    :param str name: Name of this augmenter

    >>> import nlpaug.augmenter.audio as naa
    >>> aug = naa.NoiseAug()
    """

    def __init__(self, noise_factor=0.01, name='Noise_Aug', verbose=0):
        super(NoiseAug, self).__init__(
            action=Action.SUBSTITUTE, name=name, verbose=verbose)
        self.model = self.get_model(noise_factor)

    @classmethod
    def get_model(cls, noise_factor):
        return nma.Noise(noise_factor)



from __future__ import absolute_import
from nlpaug.augmenter.audio.audio_augmenter import *
from nlpaug.augmenter.audio.noise import *
from nlpaug.augmenter.audio.shift import *
from nlpaug.augmenter.audio.speed import *
from nlpaug.augmenter.audio.pitch import *
from nlpaug.augmenter.audio.loudness import *
from nlpaug.augmenter.audio.crop import *
from nlpaug.augmenter.audio.mask import *



import numpy as np

from nlpaug.model.spectrogram import Spectrogram


class FrequencyMasking(Spectrogram):
    def __init__(self, mask_factor):
        super(FrequencyMasking, self).__init__()

        self.mask_factor = mask_factor

    def mask(self, data):
        """
            From: https://arxiv.org/pdf/1904.08779.pdf,
            Frequency masking is applied so that f consecutive mel
            frequency channels [f0, f0 + f) are masked, where f is
            first chosen from a uniform distribution from 0 to the
            frequency mask parameter F, and f0 is chosen from
            [0, v - f). v is the number of mel frequency channels.
        :return:
        """
        v  = data.shape[0]
        self.f = np.random.randint(self.mask_factor)
        self.f0 = np.random.randint(v - self.f)

        augmented_mel_spectrogram = data.copy()
        augmented_mel_spectrogram[self.f0:self.f0+self.f, :] = 0
        return augmented_mel_spectrogram



from nlpaug import Augmenter
from nlpaug.augmenter.char import CharAugmenter
from nlpaug.util import Method


class Pipeline(Augmenter, list):
    def __init__(self, action, name='Pipeline', aug_min=1, aug_p=1, flow=None, epoch=1, verbose=0):
        Augmenter.__init__(self, name=name, method=Method.FLOW,
                           action=action, aug_min=aug_min, verbose=verbose)
        self.aug_p = aug_p
        if flow is None:
            list.__init__(self, [])
        elif isinstance(flow, (Augmenter, CharAugmenter)):
            list.__init__(self, [flow])
        elif isinstance(flow, list):
            for subflow in flow:
                if not isinstance(subflow, Augmenter):
                    raise ValueError('At least one of the flow does not belongs to Augmenter')
            list.__init__(self, flow)
        else:
            raise Exception(
                'Expected None, Augmenter or list of Augmenter while {} is passed'.format(
                    type(flow)))
        self.epoch = epoch



class Action:
    INSERT = 'insert'
    SUBSTITUTE = 'substitute'
    DELETE = 'delete'
    SWAP = 'swap'

    SEQUENTIAL = 'sequential'
    SOMETIMES = 'sometimes'

    @staticmethod
    def getall():
        return [Action.INSERT, Action.SUBSTITUTE, Action.SWAP, Action.DELETE, Action.SEQUENTIAL, Action.SOMETIMES]


"""
    Augmenter that apply adjusting loudness operation to audio.
"""

from nlpaug.augmenter.audio import AudioAugmenter
from nlpaug.util import Action
import nlpaug.model.audio as nma


class LoudnessAug(AudioAugmenter):
    """
    Augmenter that crop segment of audio by random values between crop_range variable.

    :param tuple loudness_factor: Input data volume will be increased (decreased). Augmented value will be picked
            within the range of this tuple value. Volume will be reduced if value is between 0 and 1.
    :param str name: Name of this augmenter

    >>> import nlpaug.augmenter.audio as naa
    >>> aug = naa.LoudnessAug()
    """

    def __init__(self, loudness_factor=(0.5, 2), name='Loudness_Aug', verbose=0):
        super().__init__(
            action=Action.SUBSTITUTE, name=name, verbose=verbose)
        self.model = self.get_model(loudness_factor)

    @classmethod
    def get_model(cls, loudness_factor):
        return nma.Loudness(loudness_factor)



import torch
from pytorch_pretrained_bert import BertTokenizer, BertForMaskedLM

from nlpaug.model.lang_models import LanguageModels


class Bert(LanguageModels):
    START = '[CLS]'
    SEPARATOR = '[SEP]'
    MASK = '[MASK]'
    SUBWORD_PREFIX = '##'

    def __init__(self, model_path, tokenizer_path, device):
        super(Bert, self).__init__()
        self.model_path = model_path
        self.tokenizer_path = tokenizer_path
        self.device = device

        self.tokenizer = BertTokenizer.from_pretrained(tokenizer_path)
        self.model = BertForMaskedLM.from_pretrained(model_path)
        self.model.to(device)

    def predict(self, input_tokens, target_word, top_n):
        results = []

        tokens = input_tokens.copy()
        tokens.insert(0, Bert.START)
        tokens.append(Bert.SEPARATOR)

        # Mask target word
        target_pos = tokens.index(target_word)
        target_idx = self.tokenizer.convert_tokens_to_ids([target_word])[0]
        tokens[target_pos] = Bert.MASK

        # Convert feature
        token_idxes = self.tokenizer.convert_tokens_to_ids(tokens)
        segments_idxes = [0] * len(tokens)
        tokens_features = torch.tensor([token_idxes])
        segments_features = torch.tensor([segments_idxes])

        # Predict target word
        predictions = self.model(tokens_features, segments_features)

        top_score_idx = target_idx
        for _ in range(100):
            predictions[0, target_pos, top_score_idx] = -9999
            top_score_idx = torch.argmax(predictions[0, target_pos]).item()
            top_score_token = self.tokenizer.convert_ids_to_tokens([top_score_idx])[0]
            if top_score_token[:2] != Bert.SUBWORD_PREFIX:
                results.append(top_score_token)
                if len(results) >= top_n:
                    break

        return results



import re

from nlpaug.model.char import Character


class Ocr(Character):
    def __init__(self, cache=True):
        super().__init__(cache)

        self.model = self.get_model()

    def predict(self, data):
        return self.model[data]

    # TODO: Read from file
    @classmethod
    def get_model(cls):
        mapping = {
            '0': ['8', '9', 'o', 'O', 'D'],
            '1': ['4', '7', 'l', 'I'],
            '2': ['z', 'Z'],
            '5': ['8'],
            '6': ['b'],
            '8': ['s', 'S', '@', '&'],
            '9': ['g'],
            'o': ['u'],
            'r': ['k'],
            'C': ['G'],
            'O': ['D', 'U'],
            'E': ['B']
        }

        result = {}

        for k in mapping:
            result[k] = mapping[k]

        for k in mapping:
            for v in mapping[k]:
                if v not in result:
                    result[v] = []

                if k not in result[v]:
                    result[v].append(k)

        return result



"""
    Augmenter that apply spelling error simulation to textual input.
"""

import nlpaug.model.word_dict as nmwd
from nlpaug.augmenter.word import WordAugmenter
from nlpaug.util import Action

SPELLING_ERROR_MODEL = {}


def init_spelling_error_model(dict_path, include_reverse, force_reload=False):
    # Load model once at runtime
    global SPELLING_ERROR_MODEL
    if SPELLING_ERROR_MODEL and not force_reload:
        return SPELLING_ERROR_MODEL

    spelling_error_model = nmwd.Spelling(dict_path, include_reverse)

    SPELLING_ERROR_MODEL = spelling_error_model

    return SPELLING_ERROR_MODEL


class SpellingAug(WordAugmenter):
    """
    Augmenter that leverage pre-defined spelling mistake dictionary to simulate spelling mistake.

    :param str action: Either 'swap' or 'delete'. If value is 'swap', adjacent words will be swapped randomly.
        If value is 'delete', word will be removed randomly.
    :param int aug_min: Minimum number of word will be augmented.
    :param float aug_p: Percentage of word will be augmented.
    :param list stopwords: List of words which will be skipped from augment operation.
    :param func tokenizer: Customize tokenization process
    :param func reverse_tokenizer: Customize reverse of tokenization process
    :param str name: Name of this augmenter

    >>> import nlpaug.augmenter.word as naw
    >>> aug = naw.SpellingAug(dict_path='./spelling_en.txt')
    """

    def __init__(self, dict_path, name='Spelling_Aug', aug_min=1, aug_p=0.3, stopwords=None,
                 tokenizer=None, reverse_tokenizer=None, include_reverse=True, verbose=0):
        super().__init__(
            action=Action.SUBSTITUTE, name=name, aug_p=aug_p, aug_min=aug_min, stopwords=stopwords,
            tokenizer=tokenizer, reverse_tokenizer=reverse_tokenizer, verbose=verbose)

        self.dict_path = dict_path
        self.include_reverse = include_reverse
        self.model = self.get_model(force_reload=False)

    def skip_aug(self, token_idxes, tokens):
        results = []
        for token_idx in token_idxes:
            # Some words do not exit. It will be excluded in lucky draw.
            token = tokens[token_idx]
            if token in self.model.dict and len(self.model.dict[token]) > 0:
                results.append(token_idx)

        return results

    def substitute(self, data):
        results = []

        tokens = self.tokenizer(data)
        aug_idexes = self._get_aug_idxes(tokens)

        if aug_idexes is None:
            return data

        for i, token in enumerate(tokens):
            # Skip if no augment for word
            if i not in aug_idexes:
                results.append(token)
                continue

            candidate_words = self.model.predict(token)
            results.append(self.sample(candidate_words, 1)[0])

        return self.reverse_tokenizer(results)

    def get_model(self, force_reload):
        return init_spelling_error_model(self.dict_path, self.include_reverse, force_reload)



"""
    Augmenter that apply speed adjustment operation to audio.
"""

from nlpaug.augmenter.audio import AudioAugmenter
from nlpaug.util import Action
import nlpaug.model.audio as nma


class SpeedAug(AudioAugmenter):
    """
    Augmenter that crop segment of audio by random values between crop_range variable.

    :param tuple speed_range: Range of applying speed adjustment operation. Default value is (0.5, 2)
        Factor for time stretch. Audio will be slowing down if value is between 0 and 1.
        Audio will be speed up if value is larger than 1.
    :param str name: Name of this augmenter

    >>> import nlpaug.augmenter.audio as naa
    >>> aug = naa.ShiftAug()
    """

    def __init__(self, speed_range=(0.5, 2), name='Speed_Aug', verbose=0):
        super().__init__(
            action=Action.SUBSTITUTE, name=name, verbose=verbose)
        self.model = self.get_model(speed_range)

    @classmethod
    def get_model(cls, speed_range):
        return nma.Speed(speed_range)



"""
    Augmenter that apply shifting operation to audio.
"""

from nlpaug.augmenter.audio import AudioAugmenter
from nlpaug.util import Action
import nlpaug.model.audio as nma


class ShiftAug(AudioAugmenter):
    """
    Augmenter that crop segment of audio by random values between crop_range variable.

    :param int sampling_rate: sampling rate of input audio
    :param int shift_max: Max shifting segment (in second)
    :param str shift_direction: Shifting segment to left, right or one of them. Value can be 'left', 'right' or 'both'.
    :param str name: Name of this augmenter

    >>> import nlpaug.augmenter.audio as naa
    >>> aug = naa.ShiftAug(sampling_rate=44010)
    """

    def __init__(self, sampling_rate, shift_max=3, shift_direction='both', name='Shift_Aug', verbose=0):
        super(ShiftAug, self).__init__(
            action=Action.SUBSTITUTE, name=name, verbose=verbose)
        self.model = self.get_model(sampling_rate, shift_max, shift_direction)

    @classmethod
    def get_model(cls, sampling_rate, shift_max, shift_direction):
        return nma.Shift(sampling_rate, shift_max, shift_direction)


from setuptools import setup, find_packages
import sys

if sys.version_info < (3,):
    sys.exit("Sorry, Python3 is required.")

with open("README.md", encoding="utf8") as f:
    readme = f.read()

setup(
    name="nlpaug",
    version="0.0.7",
    author="Edward Ma",
    author_email="makcedward@gmail.com",
    url="https://github.com/makcedward/nlpaug",
    license="MIT",
    description="Natural language processing augmentation library for deep neural networks",
    long_description=readme,
    long_description_content_type="text/markdown",
    packages=find_packages(exclude="test"),
    keywords=[
        "deep learning", "neural network", "machine learning",
        "nlp", "natural language processing", "text", "audio", "spectrogram",
        "augmentation", "adversarial attack", "ai", "ml"]
)



"""
    Augmenter that apply cropping operation to audio.
"""

from nlpaug.augmenter.audio import AudioAugmenter
from nlpaug.util import Action
import nlpaug.model.audio as nma


class CropAug(AudioAugmenter):
    """
    Augmenter that crop segment of audio by random values between crop_range variable.

    :param int sampling_rate: sampling rate of input audio
    :param tuple crop_range: Range of applying crop operation. Default value is (0.2, 0.8)
        It means that first 20% and last 20% of data will be excluded from augment operation selection.
    :param int crop_factor: duration of cropping period (in second)
    :param str name: Name of this augmenter

    >>> import nlpaug.augmenter.audio as naa
    >>> aug = naa.CropAug(sampling_rate=44010)
    """

    def __init__(self, sampling_rate, crop_range=(0.2, 0.8), crop_factor=2, name='Crop_Aug', verbose=0):
        super().__init__(
            action=Action.DELETE, name=name, verbose=verbose)
        self.model = self.get_model(sampling_rate, crop_range=crop_range, crop_factor=crop_factor)

    def delete(self, data):
        return self.model.manipulate(data)

    @classmethod
    def get_model(cls, sampling_rate, crop_range, crop_factor):
        return nma.Corp(sampling_rate, crop_range, crop_factor)



import numpy as np

from nlpaug.model.audio import Audio


class Corp(Audio):
    def __init__(self, sampling_rate, crop_range=(0.2, 0.8), crop_factor=2):
        """

        :param sampling_rate: sampling rate of input audio
        :param crop_range: Range of applying crop operation. Default value is (0.2, 0.8)
            It means that first 20% and last 20% of data will not be excluded from augment operation. Augment operation
            will be applied to clip of rest of 60% time.
        :param crop_factor: duration of cropping period (in second)
        """
        super().__init__()
        self.sampling_rate = sampling_rate
        self.crop_range = crop_range
        self.crop_factor = crop_factor

    def manipulate(self, data):
        valid_region = (int(len(data) * self.crop_range[0]), int(len(data) * self.crop_range[1]))

        start_timeframe = np.random.randint(valid_region[0], valid_region[1])
        end_timeframe = start_timeframe + self.sampling_rate * self.crop_factor

        # Crop region is larger than valid region
        if end_timeframe > valid_region[1]:
            end_timeframe = valid_region[1]

        augmented_data = data.copy()
        augmented_data = np.delete(augmented_data, np.s_[start_timeframe:end_timeframe])

        return augmented_data



"""
    Augmenter that apply TF-IDF based to textual input.
"""

from nlpaug.augmenter.word import WordAugmenter
from nlpaug.util import Action, WarningException, WarningName, WarningCode, WarningMessage
import nlpaug.model.word_stats as nmws

TFIDF_MODEL = {}


def init_tfidf_model(model_path, force_reload=False):
    # Load model once at runtime
    global TFIDF_MODEL
    if TFIDF_MODEL and not force_reload:
        return TFIDF_MODEL

    tfidf_model = nmws.TfIdf()
    tfidf_model.read(model_path)
    TFIDF_MODEL = tfidf_model

    return tfidf_model


class TfIdfAug(WordAugmenter):
    """
    Augmenter that leverage TF-IDF statistics to insert or substitute word.

    :param str model_path: Downloaded model directory. Either model_path or model is must be provided
    :param str action: Either 'insert or 'substitute'. If value is 'insert', a new word will be injected to random
        position according to TF-IDF calculation. If value is 'substitute', word will be replaced according
        to TF-IDF calculation
    :param int aug_min: Minimum number of word will be augmented.
    :param float aug_p: Percentage of word will be augmented.
    :param list stopwords: List of words which will be skipped from augment operation.
    :param func tokenizer: Customize tokenization process
    :param func reverse_tokenizer: Customize reverse of tokenization process
    :param str name: Name of this augmenter

    >>> import nlpaug.augmenter.word as naw
    >>> aug = naw.TfIdfAug(model_path='.')
    """

    def __init__(self, model_path='.', action=Action.SUBSTITUTE,
                 name='TfIdf_Aug', aug_min=1, aug_p=0.3, aug_n=5, stopwords=None,
                 tokenizer=None, reverse_tokenizer=None, verbose=0):
        super().__init__(
            action=action, name=name, aug_p=aug_p, aug_min=aug_min, stopwords=stopwords,
            tokenizer=tokenizer, reverse_tokenizer=reverse_tokenizer, verbose=verbose)
        self.model_path = model_path
        self.aug_n = aug_n
        self.model = self.get_model(force_reload=False)

    def skip_aug(self, token_idxes, tokens):
        results = []
        for token_idx in token_idxes:
            # Some word does not come with IDF. It will be excluded in lucky draw.
            word = tokens[token_idx]

            if word in self.model.w2idf:
                results.append(token_idx)

        return results

    def _get_aug_idxes(self, tokens):
        aug_cnt = self.generate_aug_cnt(len(tokens))
        word_idxes = [i for i, t in enumerate(tokens) if self.stopwords is None or t not in self.stopwords]
        word_idxes = self.skip_aug(word_idxes, tokens)

        if len(word_idxes) == 0:
            if self.verbose > 0:
                exception = WarningException(name=WarningName.OUT_OF_VOCABULARY,
                                             code=WarningCode.WARNING_CODE_002, msg=WarningMessage.NO_WORD)
                exception.output()
            return None
        if len(word_idxes) < aug_cnt:
            aug_cnt = len(word_idxes)

        aug_probs = self.model.cal_tfidf(word_idxes, tokens)
        aug_idxes = []

        # It is possible that no token is picked. So re-try
        retry_cnt = 3
        possible_idxes = word_idxes.copy()
        for _ in range(retry_cnt):
            for i, p in zip(possible_idxes, aug_probs):
                if self.prob() < p:
                    aug_idxes.append(i)
                    possible_idxes.remove(i)

                    if len(possible_idxes) == aug_cnt:
                        break

        # If still cannot pick up, random pick index regrardless probability
        if len(aug_idxes) < aug_cnt:
            aug_idxes.extend(self.sample(possible_idxes, aug_cnt-len(aug_idxes)))

        aug_idxes = self.sample(aug_idxes, aug_cnt)

        return aug_idxes

    def insert(self, data):
        tokens = self.tokenizer(data)
        results = tokens.copy()

        aug_idxes = self._get_random_aug_idxes(tokens)
        aug_idxes.sort(reverse=True)

        for aug_idx in aug_idxes:
            original_word = results[aug_idx]
            new_word = self.sample(self.model.predict(original_word, top_n=self.aug_n), 1)[0]
            results.insert(aug_idx, new_word)

        return self.reverse_tokenizer(results)

    def substitute(self, data):
        tokens = self.tokenizer(data)
        results = tokens.copy()

        aug_idxes = self._get_aug_idxes(tokens)
        if aug_idxes is None:
            return data

        for aug_idx in aug_idxes:
            original_word = results[aug_idx]
            candidate_words = self.model.predict(original_word, top_n=self.aug_n)
            substitute_word = self.sample(candidate_words, 1)[0]

            results[aug_idx] = substitute_word

        return self.reverse_tokenizer(results)

    def get_model(self, force_reload=False):
        return init_tfidf_model(self.model_path, force_reload)




class PartOfSpeech:
    NOUN = 'noun'
    VERB = 'verb'
    ADJECTIVE = 'adjective'
    ADVERB = 'adverb'

    @staticmethod
    def wn2pos(wn_pos):
        if wn_pos == 'n':
            # Noun
            return ['NN', 'NNS', 'NNP', 'NNPS']
        if wn_pos == 'v':
            # Verb
            return ['VB', 'VBD', 'VBG', 'VBN', 'VBZ']

        if wn_pos in ['a', 's']:
            # Adjective/ Adjective Satellite
            return ['JJ', 'JJR', 'JJS', 'IN']

        if wn_pos == 'r':
            # Adverb
            return ['RB', 'RBR', 'RBS']

    @staticmethod
    def pos2wn(pos):
        if pos in ['NN', 'NNS', 'NNP', 'NNPS']:
            # Noun
            return ['n']
        elif pos in ['VB', 'VBD', 'VBG', 'VBN', 'VBZ']:
            # Verb
            return ['v']
        elif pos in ['JJ', 'JJR', 'JJS', 'IN']:
            # Adjective/ Adjective Satellite
            return ['a', 's']
        elif pos in ['RB', 'RBR', 'RBS']:
            # Adverb
            return 'r'

    @staticmethod
    def get_wn_pos():
        return [
            'NN', 'NNS', 'NNP', 'NNPS',
            'VB', 'VBD', 'VBG', 'VBN', 'VBZ',
            'JJ', 'JJR', 'JJS', 'IN',
            'RB', 'RBR', 'RBS'
        ]




class ExceptionInfo:
    def __init__(self, name, exp_type, code, msg):
        self.name = name
        self.exp_type = exp_type
        self.code = code
        self.msg = msg

    def output(self):
        msg = '[{}] Name:{}, Code:{}, Message:{}'.format(self.exp_type, self.name, self.code, self.msg)
        print(msg)


class ExceptionType:
    WARNING = 'Warning'